
#include <assert.h>
#include <stdio.h>
#include <string>
#include <vector>
#include "Board.h"

std::string Move::toString() const
{
    if (this->is_to_move()) {
        char text[3];
        text[0] = '1'+firstcell;
        text[1] = '1'+secondcell;
        text[2] = '\0';
        return text;
    } else {
        assert(this->is_to_collapse());
        if (this->collapse_into_first)
          return "collapse into first";
        else
          return "collapse into second";
    }
}

void Board::new_game()
{
    Board &b = *this;
    b.turn = X_TO_MOVE;
    b.now = 1;
    for (int i=0; i < 9; ++i)
      b.collapsed_cells[i].who = NOBODY;
    for (int i=0; i < 9; ++i)
      for (int j=0; j < 9; ++j)
        b.adj[i][j].who = NOBODY;
}

bool Board::cell_is_collapsed(int c) const
{
    return (collapsed_cells[c].who != NOBODY);
}

bool Board::is_valid_move(const Move &m) const
{
    const Board &b = *this;
    if (b.turn != m.kind) return false;
    if (b.is_to_move()) {
        if (m.firstcell < 0 || 8 < m.firstcell) return false;
        if (m.secondcell < 0 || 8 < m.secondcell) return false;
        if (m.firstcell == m.secondcell) return false;
        if (b.cell_is_collapsed(m.firstcell)) return false;
        if (b.cell_is_collapsed(m.secondcell)) return false;
        return true;
    } else {
        assert(b.is_to_collapse());
        /* Collapsing left or collapsing right are the only possibilities,
         * and they're both always valid moves. */
        return true;
    }
}


static bool find_win(const Board &b, int x, int y, int z, int &xfirst, int &ofirst)
{
    Player who = b.collapsed_cells[x].who;
    if (who == NOBODY) return false;
    assert(b.cell_is_collapsed(x));
    if (b.collapsed_cells[y].who != who) return false;
    assert(b.cell_is_collapsed(y));
    if (b.collapsed_cells[z].who != who) return false;
    assert(b.cell_is_collapsed(z));
    /* This is a three-in-a-row for somebody! */
    int completed_at = b.collapsed_cells[x].when;
    completed_at = std::max(completed_at, b.collapsed_cells[y].when);
    completed_at = std::max(completed_at, b.collapsed_cells[z].when);
    assert(1 <= completed_at && completed_at <= b.now);
    if (who == 'X') {
        xfirst = std::min(xfirst, completed_at);
    } else {
        assert(who == 'O');
        ofirst = std::min(ofirst, completed_at);
    }
    return true;
}

Player Board::winner() const
{
    const Board &b = *this;
    if (b.is_to_collapse())
      return NOBODY;
    int xfirst = 999;
    int ofirst = 999;
    bool somebody_won = false;
    somebody_won |= find_win(b, 0,1,2, xfirst, ofirst);
    somebody_won |= find_win(b, 3,4,5, xfirst, ofirst);
    somebody_won |= find_win(b, 6,7,8, xfirst, ofirst);
    somebody_won |= find_win(b, 0,3,6, xfirst, ofirst);
    somebody_won |= find_win(b, 1,4,7, xfirst, ofirst);
    somebody_won |= find_win(b, 2,5,8, xfirst, ofirst);
    somebody_won |= find_win(b, 0,4,8, xfirst, ofirst);
    somebody_won |= find_win(b, 2,4,6, xfirst, ofirst);
    if (somebody_won)
      return ((xfirst < ofirst) ? 'X' : 'O');
    return NOBODY;
}

bool Board::game_over() const
{
    if (this->is_to_collapse()) return false;
    std::vector<Move> allmoves;
    findmoves(*this, allmoves);
    return allmoves.empty();
}


/* Return true if cells "a" and "b" are connected via a chain of
 * entanglements. */
bool Board::path_exists_between(int a, int b) const
{
    assert(a != b);
    assert(!cell_is_collapsed(a));

    assert(!cell_is_collapsed(b));
    bool visited[9] = {0};
    bool frontier[9] = {0};
    frontier[a] = true;
    while (true) {
        /* Pick the first frontier cell. */
        int f;
        for (f = 0; f < 9 && !frontier[f]; ++f) { }
        /* If the frontier is empty, we're done. */
        if (f == 9) return false;
        if (f == b) return true;
        /* Visit this frontier cell by adding all its neighbors
         * to the frontier list. */
        for (int n = 0; n < 9; ++n) {
            if (!visited[n] && (this->adj[f][n].who != NOBODY))
              frontier[n] = true;
        }
        /* Go back and get the next frontier cell to visit. */
        frontier[f] = false;
        visited[f] = true;
    }
    /*NOTREACHED*/
    assert(false);
    return false;
}


/* Collapse cell c to (who,when). Then collapse any dependencies. */
static void try_collapse(Board &b, int c, Player who, int when)
{
    assert(!b.cell_is_collapsed(c));
    assert(who == 'X' || who == 'O');
    b.collapsed_cells[c].who = who;
    b.collapsed_cells[c].when = when;
    for (int c2 = 0; c2 < 9; ++c2) {
        if (b.adj[c][c2].who != NOBODY) {
            /* There is an edge between c and c2. Collapse it. */
            assert(b.adj[c][c2] == b.adj[c2][c]);
            Player c2who = b.adj[c][c2].who;
            int c2when = b.adj[c][c2].when;
            b.adj[c][c2].who = NOBODY;
            b.adj[c2][c].who = NOBODY;
            try_collapse(b, c2, c2who, c2when);
        }
    }
}

void applymove(Board &b, const Move &m)
{
    assert(b.is_valid_move(m));
    if (b.is_to_move()) {
        if (b.path_exists_between(m.firstcell, m.secondcell)) {
            b.last_move = m;
            b.turn = ((b.turn == X_TO_MOVE) ? O_TO_COLLAPSE : X_TO_COLLAPSE);
        } else {
            b.adj[m.firstcell][m.secondcell].who = b.attacker();
            b.adj[m.firstcell][m.secondcell].when = b.now;
            b.adj[m.secondcell][m.firstcell] =
                    b.adj[m.firstcell][m.secondcell];
            b.turn = ((b.turn == X_TO_MOVE) ? O_TO_MOVE : X_TO_MOVE);
            b.now += 1;
        }
    } else {
        assert(b.is_to_collapse());
        assert(b.last_move.kind == (b.attacker()=='X' ? O_TO_MOVE : X_TO_MOVE));
        const int firstcell = b.last_move.firstcell;
        const int secondcell = b.last_move.secondcell;
        assert(!b.cell_is_collapsed(firstcell));
        assert(!b.cell_is_collapsed(secondcell));
        if (b.adj[firstcell][secondcell].who != NOBODY) {
            /* Special case for a 2-cycle. */
            assert(b.adj[firstcell][secondcell] == b.adj[secondcell][firstcell]);
            if (m.collapse_into_first) {
                try_collapse(b, firstcell, b.defender(), b.now);
            } else {
                try_collapse(b, secondcell, b.defender(), b.now);
            }
            b.adj[firstcell][secondcell].who = NOBODY;
            b.adj[secondcell][firstcell].who = NOBODY;
        } else {
            if (m.collapse_into_first) {
                try_collapse(b, firstcell, b.defender(), b.now);
            } else {
                try_collapse(b, secondcell, b.defender(), b.now);
            }
        }
        b.turn = ((b.turn == X_TO_COLLAPSE) ? X_TO_MOVE : O_TO_MOVE);
        b.now += 1;
    }
}

void findmoves(const Board &b, std::vector<Move> &allmoves)
{
    if (b.winner() != NOBODY) return;
    if (b.is_to_collapse()) {
        Move m;
        m.kind = b.turn;
        m.collapse_into_first = true;
        allmoves.push_back(m);
        m.collapse_into_first = false;
        allmoves.push_back(m);
    } else {
        Move m;
        m.kind = b.turn;
        /* Pick any two uncollapsed cells. */
        for (int i=0; i < 9; ++i) {
            if (b.cell_is_collapsed(i)) continue;
            m.secondcell = i;
            for (int j=0; j < i; ++j) {
                if (b.cell_is_collapsed(j)) continue;
                m.firstcell = j;
                allmoves.push_back(m);
            }
        }
    }
}

int findattacker(const Board &b)
{
    return ((b.attacker() == 'X') ? 0 : 1);
}


void Board::print(FILE *out) const
{
    const Board &b = *this;
    char text[9][18];
    memset(text, ' ', sizeof text);
    for (int i=0; i < 9; ++i) {
        for (int j=0; j < 9; ++j) {
            if (b.adj[i][j].who == NOBODY) continue;
            text[3*(i/3)+(j/3)][2*(3*(i%3)+(j%3))+0] = b.adj[i][j].who;
            text[3*(i/3)+(j/3)][2*(3*(i%3)+(j%3))+1] =
                    "0123456789abcdefghjiklmnopqrstuvwxyz"[b.adj[i][j].when];
        }
    }
    for (int i=0; i < 9; ++i) {
        if (!b.cell_is_collapsed(i)) continue;
        text[3*(i/3)+0][6*(i%3)+0] = '/';
        text[3*(i/3)+0][6*(i%3)+5] = '\\';
        text[3*(i/3)+1][6*(i%3)+2] = b.collapsed_cells[i].who;
        text[3*(i/3)+1][6*(i%3)+3] =
                    "0123456789abcdefghjiklmnopqrstuvwxyz"[b.collapsed_cells[i].when];
        text[3*(i/3)+2][6*(i%3)+0] = '\\';
        text[3*(i/3)+2][6*(i%3)+5] = '/';
    }
    fprintf(out, "%.6s|%.6s|%.6s\n", text[0], text[0]+6, text[0]+12);
    fprintf(out, "%.6s|%.6s|%.6s\n", text[1], text[1]+6, text[1]+12);
    fprintf(out, "%.6s|%.6s|%.6s\n", text[2], text[2]+6, text[2]+12);
    fprintf(out, "------+------+------\n");
    fprintf(out, "%.6s|%.6s|%.6s\n", text[3], text[3]+6, text[3]+12);
    fprintf(out, "%.6s|%.6s|%.6s\n", text[4], text[4]+6, text[4]+12);
    fprintf(out, "%.6s|%.6s|%.6s\n", text[5], text[5]+6, text[5]+12);
    fprintf(out, "------+------+------\n");
    fprintf(out, "%.6s|%.6s|%.6s\n", text[6], text[6]+6, text[6]+12);
    fprintf(out, "%.6s|%.6s|%.6s\n", text[7], text[7]+6, text[7]+12);
    fprintf(out, "%.6s|%.6s|%.6s\n", text[8], text[8]+6, text[8]+12);
}


